import logging
import logging.config
from pathlib import Path


mylogger = logging.getLogger('zkpytb.logging')


def setup_simple_console_and_file_logger(logger_name, logfile=True,
                                         logdir=None, logfilename=None,
                                         log_level='DEBUG', options=None):
    """
    TODOC
    """

    if options is None:
        options = {}

    # default values
    do_file_logging = True
    log_directory = Path('.')
    log_filename = logger_name + '.log'

    if logdir is not None:
        try:
            log_directory = Path(logdir)
        except TypeError:
            mylogger.exception('Invalid type for argument "logdir". '
                               'Expected "str", "bytes" or "pathlib.Path". '
                               'Received type: {}'.format(type(logdir)))
            do_file_logging = False
        else:
            if not log_directory.is_dir():
                mylogger.warning('The given logdir is not a directory!')
                do_file_logging = False

    if not logfile:
        do_file_logging = False

    if logfilename is not None:
        log_filename = logfilename

    log_filepath = log_directory / log_filename

    default_format_str = '{asctime},{msecs:0>3.0f} - {levelname:8} - {message}'
    format_str = options.get('format_str', default_format_str)

    default_datefmt_str = '%Y-%m-%d %H:%M:%S'
    datefmt_str = options.get('datefmt_str', default_datefmt_str)

    console_formatter = options.get('console_formatter', 'normal_extra_newlines')
    file_formatter = options.get('file_formatter', 'normal')

    console_handler = {
        'level': log_level,
        'class': 'logging.StreamHandler',
        'formatter': console_formatter
    }

    file_handler = {
        'level': log_level,
        'class': 'logging.FileHandler',
        'formatter': file_formatter,
        'filename': str(log_filepath),
    }

    logging_conf = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'normal': {
                'format': format_str,
                'datefmt': datefmt_str,
                'style': '{'
            },
            'normal_extra_newlines': {
                'format': format_str + '\n',
                'datefmt': datefmt_str,
                'style': '{'
            },
        },
        'handlers': {
            'console': console_handler
        },
        'loggers': {
            logger_name: {
                'handlers': ['console'],
                'level': log_level,
            },
        }
    }

    if do_file_logging:
        logging_conf['handlers']['file'] = file_handler
        logging_conf['loggers'][logger_name]['handlers'].append('file')

    logging.config.dictConfig(logging_conf)
    logger = logging.getLogger(logger_name)

    return logger



"""
Miscellaneous helper functions

Author: Marc Gallet
Date: 2017-06
"""

import hashlib
import logging
import subprocess


mylogger = logging.getLogger('zkpytb.utils')


def hashfile(filepath, hash_method='sha256', BLOCKSIZE=65536):
    """Hash a file"""

    hasher = hashlib.new(hash_method)

    with open(str(filepath), 'rb') as afile:
        buf = afile.read(BLOCKSIZE)
        while len(buf) > 0:
            hasher.update(buf)
            buf = afile.read(BLOCKSIZE)

    return hasher.hexdigest()


def hashstring(inputstring, hash_method='sha256'):
    """Hash a file"""

    hasher = hashlib.new(hash_method)
    hasher.update(inputstring)
    return hasher.hexdigest()


def get_git_hash(rev='HEAD'):
    """Get the git hash of the current directory"""

    git_hash = ''
    try:
        git_out = subprocess.check_output(['git', 'rev-parse', rev], universal_newlines=True)
    except subprocess.CalledProcessError:
        mylogger.exception("Couldn't determine the git hash!")
    else:
        git_hash = git_out.strip()

    return git_hash



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `utils` module of `zkpytb` package."""

import os
import pytest
import subprocess


from zkpytb.utils import (
    hashfile,
    hashstring,
    get_git_hash,
)


@pytest.fixture(scope='module',
                params='sha512,sha256,sha1,md5'.split(','))
def a_hash_method(request):
    return request.param


def test_hashstring_hashmethods(a_hash_method):
    hash_res1 = hashstring(b'', hash_method=a_hash_method)
    hash_res2 = hashstring(b'test', hash_method=a_hash_method)
    assert isinstance(hash_res1, str)
    assert len(hash_res1) > 0
    assert hash_res1 != hash_res2


def test_hashfile_hashmethods(a_hash_method, tmpdir):
    f1 = tmpdir.join("hashfile1.txt")
    f2 = tmpdir.join("hashfile2.txt")
    f1.write("content1")
    f2.write("content2")
    hashfile_res1 = hashfile(f1, hash_method=a_hash_method)
    hashfile_res2 = hashfile(f2, hash_method=a_hash_method)
    assert isinstance(hashfile_res1, str)
    assert len(hashfile_res1) > 0
    assert hashfile_res1 != hashfile_res2


@pytest.mark.xfail(reason="Tox temp dir is located under the main git repository of the project...")
def test_get_git_hash_nogit(tmpdir):
    curdir = os.getcwd()
    try:
        nogit_dir = tmpdir.mkdir("nogit_dir")
        os.chdir(str(nogit_dir))
        git_hash = get_git_hash()
    finally:
        os.chdir(curdir)

    assert git_hash == ''


def test_get_git_hash_emptygit(tmpdir):
    curdir = os.getcwd()
    try:
        emptygit_dir = tmpdir.mkdir("emptygit_dir")
        os.chdir(str(emptygit_dir))
        subprocess.check_call(['git', 'init'])
        git_hash = get_git_hash()
    finally:
        os.chdir(curdir)

    assert git_hash == ''


def test_get_git_hash_minimalgit(tmpdir):
    curdir = os.getcwd()
    try:
        minimalgit_dir = tmpdir.mkdir("minimalgit_dir")
        minimalgit_dir.join('testfile.txt').write('minimalgit')
        os.chdir(str(minimalgit_dir))
        subprocess.check_call(['git', 'init'])
        subprocess.check_call(['git', 'config', 'user.email', 'test@domain.invalid'])
        subprocess.check_call(['git', 'config', 'user.name', 'Test User'])
        subprocess.check_call(['git', 'add', 'testfile.txt'])
        subprocess.check_call(['git', 'commit', '-m', 'testcommit'])
        git_hash = get_git_hash()
    finally:
        os.chdir(curdir)

    assert isinstance(git_hash, str)
    assert len(git_hash) == 40



"""
An implementation of a priority queue based on heapq and
https://docs.python.org/3/library/heapq.html#priority-queue-implementation-notes

Author: Marc Gallet
Date: 2018-01
"""

import heapq
import itertools


class EmptyQueueError(Exception):
    pass


class PriorityQueue:
    """Based on https://docs.python.org/3/library/heapq.html#priority-queue-implementation-notes"""
    def __init__(self, name=''):
        self.name = name
        self.pq = []                         # list of entries arranged in a heap
        self.entry_finder = {}               # mapping of tasks to entries
        self.counter = itertools.count()     # unique sequence count
        self.num_tasks = 0                  # track the number of tasks in the queue

    def __len__(self):
        return self.num_tasks

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.pop_task()
        except EmptyQueueError:
            raise StopIteration

    @property
    def empty(self):
        return self.num_tasks == 0

    def add_task(self, task, priority=0):
        'Add a new task or update the priority of an existing task'
        if task in self.entry_finder:
            self.remove_task(task)
        count = next(self.counter)
        removed = False
        entry = [priority, count, task, removed]
        self.entry_finder[task] = entry
        heapq.heappush(self.pq, entry)
        self.num_tasks += 1

    def remove_task(self, task):
        'Mark an existing task as REMOVED. Raise KeyError if not found.'
        entry = self.entry_finder.pop(task)
        entry[-1] = True  # mark the element as removed
        self.num_tasks -= 1

    def pop_task(self):
        'Remove and return the lowest priority task. Raise KeyError if empty.'
        while self.pq:
            priority, count, task, removed = heapq.heappop(self.pq)
            if not removed:
                del self.entry_finder[task]
                self.num_tasks -= 1
                return task
        raise EmptyQueueError('pop from an empty priority queue')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `logging` module of `zkpytb` package."""

import logging
import pytest
from pathlib import Path
# from testfixtures import Comparison as C, compare
from testfixtures import LogCapture

from zkpytb.logging import setup_simple_console_and_file_logger


loglevels = 'CRITICAL,ERROR,WARNING,INFO,DEBUG'.split(',')

loglevel_ranks = {
    'CRITICAL': 1,
    'ERROR': 2,
    'WARNING': 3,
    'INFO': 4,
    'DEBUG': 5,
}


@pytest.fixture(scope='module',
                params=loglevels)
def a_log_level(request):
    return request.param


def log_some_messages(logger, level='DEBUG'):
    max_rank = loglevel_ranks[level]
    if max_rank >= 1:
        logger.critical('critical message')
    if max_rank >= 2:
        logger.error('error message')
    if max_rank >= 3:
        logger.warning('warning message')
    if max_rank >= 4:
        logger.info('info message')
    if max_rank >= 5:
        logger.debug('debug message')


def expected_some_messages(logger_name, level='DEBUG'):
    expected_messages = []
    max_rank = loglevel_ranks[level]
    if max_rank >= 1:
        expected_messages.append((logger_name, 'CRITICAL', 'critical message'))
    if max_rank >= 2:
        expected_messages.append((logger_name, 'ERROR', 'error message'))
    if max_rank >= 3:
        expected_messages.append((logger_name, 'WARNING', 'warning message'))
    if max_rank >= 4:
        expected_messages.append((logger_name, 'INFO', 'info message'))
    if max_rank >= 5:
        expected_messages.append((logger_name, 'DEBUG', 'debug message'))
    return expected_messages


def test_setup_simple_console_and_file_logger_1():
    """
    Test the case without file logging.
    """
    logger_name = 'logtest1'
    with LogCapture('zkpytb.logging') as lc2:
        logger = setup_simple_console_and_file_logger(logger_name, logfile=False)
    assert len(lc2.records) == 0
    assert len(logger.handlers) == 1
    assert logger.level == logging.DEBUG
    with LogCapture(logger_name) as lc1:
        log_some_messages(logger)
    lc1.check(*expected_some_messages(logger_name))
    log_filename = logger_name + '.log'
    log_filepath = Path('.') / log_filename
    assert not log_filepath.exists()


def test_setup_simple_console_and_file_logger_2(tmpdir):
    """
    Test the case with file logging.
    """
    logger_name = 'logtest2'
    with LogCapture('zkpytb.logging') as lc2:
        logger = setup_simple_console_and_file_logger(logger_name, logdir=str(tmpdir))
    assert len(lc2.records) == 0
    assert len(logger.handlers) == 2
    log_some_messages(logger)
    log_filename = logger_name + '.log'
    log_filepath = Path(str(tmpdir)) / log_filename
    assert log_filepath.is_file()
    with open(str(log_filepath), 'r') as f:
        lines = f.readlines()
    assert len(lines) == 5
    for i, logtuple in enumerate(expected_some_messages(logger_name)):
        assert logtuple[1] in lines[i]
        assert logtuple[2] in lines[i]


def test_setup_simple_console_and_file_logger_3(tmpdir):
    """
    Test the case where the log directory does not exists.
    """
    logger_name = 'logtest3'
    with LogCapture('zkpytb.logging') as lc2:
        logger = setup_simple_console_and_file_logger(logger_name, logdir=str(tmpdir) + '_nonexistent')
    lc2.check(('zkpytb.logging', 'WARNING', 'The given logdir is not a directory!'))
    assert len(logger.handlers) == 1
    with LogCapture(logger_name) as lc1:
        log_some_messages(logger)
    lc1.check(*expected_some_messages(logger_name))
    log_filename = logger_name + '.log'
    log_filepath = Path(str(tmpdir)) / log_filename
    assert not log_filepath.exists()


def test_setup_simple_console_and_file_logger_4(tmpdir):
    """
    Test the case where the logdir is of invalid type.
    """
    logger_name = 'logtest4'
    with LogCapture('zkpytb.logging') as lc2:
        logger = setup_simple_console_and_file_logger(logger_name, logdir=123)
    err_msg = ('Invalid type for argument "logdir". '
               'Expected "str", "bytes" or "pathlib.Path". '
               'Received type: {}'.format(type(123)))
    lc2.check(('zkpytb.logging', 'ERROR', err_msg))
    assert len(logger.handlers) == 1
    with LogCapture(logger_name) as lc1:
        log_some_messages(logger)
    lc1.check(*expected_some_messages(logger_name))
    log_filename = logger_name + '.log'
    log_filepath = Path('.') / log_filename
    assert not log_filepath.exists()


def test_setup_simple_console_and_file_logger_5(tmpdir):
    """
    Test the case where the logfile name is supplied by the user.
    """
    logfilename = 'loremipsum.log'
    logger_name = 'logtest5'
    with LogCapture('zkpytb.logging') as lc2:
        logger = setup_simple_console_and_file_logger(logger_name, logdir=str(tmpdir),
                                                      logfilename=logfilename)
    assert len(lc2.records) == 0
    assert len(logger.handlers) == 2
    log_some_messages(logger)
    log_filepath = Path(str(tmpdir)) / logfilename
    assert log_filepath.is_file()
    with open(str(log_filepath), 'r') as f:
        lines = f.readlines()
    assert len(lines) == 5
    for i, logtuple in enumerate(expected_some_messages(logger_name)):
        assert logtuple[1] in lines[i]
        assert logtuple[2] in lines[i]


def test_setup_simple_console_and_file_logger_6(tmpdir):
    """
    Test the case with extra options and at the same time the case with different file formatting.
    """
    logger_name = 'logtest6'
    options = {'file_formatter': 'normal_extra_newlines'}
    with LogCapture('zkpytb.logging') as lc2:
        logger = setup_simple_console_and_file_logger(logger_name, logdir=str(tmpdir), options=options)
    assert len(lc2.records) == 0
    assert len(logger.handlers) == 2
    log_some_messages(logger)
    log_filename = logger_name + '.log'
    log_filepath = Path(str(tmpdir)) / log_filename
    assert log_filepath.is_file()
    with open(str(log_filepath), 'r') as f:
        lines = f.readlines()
    assert len(lines) == 10
    for i, logtuple in enumerate(expected_some_messages(logger_name)):
        assert logtuple[1] in lines[2 * i]
        assert logtuple[2] in lines[2 * i]


def test_setup_simple_console_and_file_logger_7(a_log_level):
    """
    Test the case without file logging and with different log levels.
    """
    logger_name = 'logtest7'
    with LogCapture('zkpytb.logging') as lc2:
        logger = setup_simple_console_and_file_logger(logger_name, logfile=False, log_level=a_log_level)
    assert len(lc2.records) == 0
    assert len(logger.handlers) == 1
    assert logger.level == getattr(logging, a_log_level)
    with LogCapture(logger_name) as lc1:
        log_some_messages(logger, level=a_log_level)
    lc1.check(*expected_some_messages(logger_name, level=a_log_level))
    log_filename = logger_name + '.log'
    log_filepath = Path('.') / log_filename
    assert not log_filepath.exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `dicts` module of `zkpytb` package."""

import pytest
import sys


from zkpytb.dicts import (
    AutoDict,
    AutoOrderedDict,
    filter_dict_callfunc,
    filter_dict_only_scalar_values,
    filter_dict_with_keylist,
    mergedicts,
    dict_stable_json_repr,
    hashdict,
    dict_values_map,
)


def dummyfunc():
    return


@pytest.fixture(scope='module',
                params=[None, 0, 1, 3.14, set(), 'string', True, dict(), list()],
                ids=['none', '0', '1', 'float', 'set', 'string', 'bool', 'dict', 'list'])
def not_a_callable(request):
    return request.param


@pytest.fixture(scope='module',
                params=[None, 0, 1, 3.14, set(), 'string', True, list(), dummyfunc],
                ids=['none', '0', '1', 'float', 'set', 'string', 'bool', 'list', 'func'])
def not_a_dict(request):
    return request.param


@pytest.fixture(scope='module',
                params=[None, 0, 1, 3.14, set(), 'string', True, dict(), dummyfunc],
                ids=['none', '0', '1', 'float', 'set', 'string', 'bool', 'dict', 'func'])
def not_a_list(request):
    return request.param


@pytest.fixture(scope='module')
def dict1():
    return {
        'a': 1,
        'b': 2,
        'c': [7, 8],
        'd': (4, 5),
        'e': {'e1': None, 'e2': 'test'},
        'f': dummyfunc,
        'n': None,
    }


@pytest.fixture(scope='module')
def dict2():
    return {
        'e': {'e2': 'test2', 'e3': {'e3a': 0, 'e3b': [None]}},
        'x': 'xxx',
    }


@pytest.fixture(scope='module')
def dict3():
    return {
        '1%': 181,
        '25%': 321,
        '5%': 243,
        '50%': 647,
        '75%': 1675,
        '95%': 3177,
        '99%': 3278,
        'count': 43243,
        'iqr': 1354,
        'mad': 523,
        'mad_c1': 353,
        'max': 3357,
        'mean': 1117,
        'min': 167,
        'std': 1045,
    }


def test_filter_dict_callfunc_not_a_dict(not_a_dict):
    with pytest.raises(AssertionError):
        filter_dict_callfunc(not_a_dict, dummyfunc)


def test_filter_dict_callfunc_not_a_callable(not_a_callable):
    with pytest.raises(AssertionError):
        filter_dict_callfunc({}, not_a_callable)


def test_filter_dict_callfunc(dict1):
    res = filter_dict_callfunc(dict1, lambda k, v: k in 'abfn')
    assert sorted(list(res.keys())) == ['a', 'b', 'f', 'n']
    res = filter_dict_callfunc(dict1, lambda k, v: isinstance(v, (list, dict)))
    assert sorted(list(res.keys())) == ['c', 'e']


def test_filter_dict_only_scalar_values_not_a_dict(not_a_dict):
    with pytest.raises(AssertionError):
        filter_dict_only_scalar_values(not_a_dict)


def test_filter_dict_only_scalar_values(dict1):
    res = filter_dict_only_scalar_values(dict1)
    assert sorted(list(res.keys())) == ['a', 'b', 'f', 'n']


def test_filter_dict_with_keylist_not_a_dict(not_a_dict):
    with pytest.raises(AssertionError):
        filter_dict_with_keylist(not_a_dict, [])


def test_filter_dict_with_keylist_not_a_list(not_a_list):
    with pytest.raises(AssertionError):
        filter_dict_with_keylist({}, not_a_list)


def test_filter_dict_with_keylist(dict1):
    res = filter_dict_with_keylist(dict1, ['a', 'd'])
    assert sorted(list(res.keys())) == ['a', 'd']


def test_filter_dict_with_keylist_blacklist(dict1):
    blacklist = ['a', 'd']
    orig_keys = set(dict1.keys())
    res = filter_dict_with_keylist(dict1, blacklist, blacklistmode=True)
    res_keys = set(res.keys())
    assert res_keys == orig_keys - set(blacklist)


def test_mergedicts_1(dict1, dict2):
    res = dict(mergedicts(dict1, dict2))
    expected_res = {
        'a': 1,
        'b': 2,
        'c': [7, 8],
        'd': (4, 5),
        'e': {'e1': None, 'e2': 'test2', 'e3': {'e3a': 0, 'e3b': [None]}},
        'f': dummyfunc,
        'n': None,
        'x': 'xxx',
    }
    assert res == expected_res


def test_mergedicts_not_a_dict(dict1, not_a_dict):
    with pytest.raises(AssertionError):
        dict(mergedicts(dict1, not_a_dict))


def test_hashdict_not_a_dict(not_a_dict):
    with pytest.raises(AssertionError):
        hashdict(not_a_dict)


def test_hashdict_1(dict2):
    py36 = sys.version_info >= (3, 6)
    res = {}
    res['default'] = hashdict(dict2)
    methods = ['sha1', 'sha256', 'sha512', 'md5']
    if py36:
        methods += ['blake2b', 'blake2s']
    for m in methods:
        res[m] = hashdict(dict2, method=m)
    assert res['sha1'] == '5fa3c59b2295cdb4c7a01c20e086971a0bcfcb7d'
    assert res['sha256'] == 'ca7b51eb5b04a2c8aee95b7e731d971ecce750b0bd24ab8c90156e3b1f22aeaa'
    assert res['sha512'] == ('f7fcc1885ef531ba1c4e0e56b3c38f5cd398cbef1778ee404559d6cd34083c32'
                             '88c63dd9fee4834fbea0bc2b8f04cb66309d4468ba55ad47c76961cd293e20dc')
    assert res['md5'] == 'e1926c486437ca20489d4c35210db768'
    if py36:
        assert res['blake2b'] == ('01303535c6f0b50006d31d36843489d8ec9364a60f85a2802e53676fa78fd671'
                                  '9251664cd7dea5af3f46f7b6625b35e55c71863e99e624d687bbb32440bd573d')
        assert res['blake2s'] == '7b418c386652a926453cefc8fcc0710a13e527c98980df2295c73a11c4bb9374'
    assert res['default'] == res['sha1']


def test_hashdict_2():
    import datetime
    import decimal
    import pathlib
    import uuid
    d = {
        'None': None,
        'bool': True,
        'int': 123456789,
        'complex': 1 + 2j,
        'float': 3.14159,
        'infinity': float('inf'),
        'nan': float('nan'),
        'string': 'lalala',
        'decimal': decimal.Decimal('1.1618'),
        'pathlib.Path': pathlib.Path('test/a/path.test'),
        'datetime.timedelta': datetime.timedelta(days=1, minutes=2, seconds=3),
        'datetime.date': datetime.date(2007, 12, 5),
        'datetime.time': datetime.time(17, 28, 59),
        'datetime.datetime': datetime.datetime(2007, 12, 5, 17, 28, 59),
        'uuid': uuid.UUID('{00010203-0405-0607-0809-0a0b0c0d0e0f}'),
    }
    d_json_repr = dict_stable_json_repr(d)
    assert d_json_repr == (
        '''{"None": null, "bool": true, "complex": "(1+2j)", "datetime.date": "2007-12-05",'''
        ''' "datetime.datetime": "2007-12-05T17:28:59", "datetime.time": "17:28:59", "datetime.timedelta": 86523.0,'''
        ''' "decimal": "1.1618", "float": 3.14159, "infinity": Infinity, "int": 123456789, "nan": NaN,'''
        ''' "pathlib.Path": "test/a/path.test", "string": "lalala", "uuid": "00010203-0405-0607-0809-0a0b0c0d0e0f"}'''
    )
    res = hashdict(d)
    assert res == '1aa572054769b78af1e9ed7cbd70cc1cb149a90b'


def test_hashdict_3():
    with pytest.raises(TypeError):
        hashdict({'lambda': lambda x: x})


def test_autodict(base_class=AutoDict):
    adict = base_class()
    assert isinstance(adict, dict)
    assert isinstance(adict['x'], base_class)
    assert isinstance(adict['x']['y'], base_class)
    assert isinstance(adict['x']['y']['z'], base_class)
    adict['a'] = 123
    assert adict['a'] == 123
    adict['b']['b1'] = 456
    assert adict['b']['b1'] == 456
    adict['b']['b2'] = 789
    assert adict['b']['b2'] == 789
    adict['c']['c1']['c11'] = 999
    assert adict['c']['c1']['c11'] == 999
    adict['d'] = {}
    with pytest.raises(KeyError):
        adict['d']['d1']['d2'] = None
    adict_as_dict = adict.to_dict()
    assert not isinstance(adict_as_dict, base_class)
    assert not isinstance(adict_as_dict['b'], base_class)
    with pytest.raises(KeyError):
        adict_as_dict['e']['e1']['e2'] = None
    if isinstance(adict, AutoOrderedDict):
        assert list(adict.keys()) == ['x', 'a', 'b', 'c', 'd']


def test_autoordereddict():
    test_autodict(base_class=AutoOrderedDict)


def test_dict_values_map_1(dict3):
    res = dict_values_map(lambda x: x % 10, dict3)
    expected_res = {
        '1%': 1,
        '25%': 1,
        '5%': 3,
        '50%': 7,
        '75%': 5,
        '95%': 7,
        '99%': 8,
        'count': 3,
        'iqr': 4,
        'mad': 3,
        'mad_c1': 3,
        'max': 7,
        'mean': 7,
        'min': 7,
        'std': 5,
    }
    assert res == expected_res
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `utils` module of `zkpytb` package."""

import pytest

from zkpytb.priorityqueue import (
    EmptyQueueError,
    PriorityQueue,
)


def test_priority_queue_add_pop_task():
    pq = PriorityQueue()
    assert len(pq) == 0
    assert pq.empty
    tasks = [
        (999, 'last'),
        (10, 'to be replaced'),
        (5, 'middle'),
        (5, 'to be removed'),
        (1, 'first'),
        (5, 'middle2'),
        (-1, 'before first'),
    ]
    for prio, task in tasks:
        pq.add_task(task, prio)
    assert not pq.empty
    assert len(pq) == 7
    assert pq.entry_finder['to be replaced'][0] == 10
    pq.add_task('to be replaced', 88)
    assert len(pq) == 7
    assert pq.entry_finder['to be replaced'][0] == 88
    pq.remove_task('to be removed')
    assert len(pq) == 6
    assert pq.pop_task() == 'before first'
    assert pq.pop_task() == 'first'
    assert pq.pop_task() == 'middle'
    assert pq.pop_task() == 'middle2'
    assert pq.pop_task() == 'to be replaced'
    assert pq.pop_task() == 'last'
    assert len(pq) == 0
    with pytest.raises(EmptyQueueError) as excinfo:
        pq.pop_task()
    assert excinfo.value.args[0] == 'pop from an empty priority queue'
    assert pq.empty


def test_priority_queue_iter():
    pq = PriorityQueue(name='test_iter')
    assert pq.name == 'test_iter'
    tasks = [
        (999, 'last'),
        (5, 'middle'),
        (1, 'first'),
        (5, 'middle2'),
        (-1, 'before first'),
    ]
    for prio, task in tasks:
        pq.add_task(task, prio)
    expected_tasks = [
        'before first',
        'first',
        'middle',
        'middle2',
        'last',
    ]
    i = 0
    for task in pq:
        assert task == expected_tasks[i]
        i += 1



"""
Small helper functions related to pandas functionalities

Author: Marc Gallet (2017)
"""

try:
    import numpy as np
    import pandas as pd
    import scipy
except ImportError:  # pragma: no cover
    raise ImportError(
        'numpy, pandas and scipy packages are required in order to use this module '
        'but they will not installed automatically by the zkpytb package. '
        'Please install them yourself.'
    )


# More percentiles when using pd.describe()
extended_percentiles = [.01, .05, .25, .5, .75, .95, .99]


def tdescr(df_in, percentiles=None, disp=True):
    """
    Helper function to display and return the transposition
    of the output of DataFrame.describe(). This means that
    the columns of the input DataFrame are returned as rows
    and the various statistical values of interest are the
    columns in the result.

    It is useful when the input dataframe has a lot of columns.

    This function additionally computes the statistics on a
    larger set of percentiles than the default DataFrame.describe()
    function.
    """
    try:
        from IPython.display import display
    except ImportError:
        display = print

    if not percentiles:
        percentiles = [.01, .05, .25, .5, .75, .95, .99]

    tdescr_out = df_in.describe(percentiles).T
    if disp:
        display(tdescr_out)
    return tdescr_out


def df_query_with_ratio(df_in, query, ratio_name='ratio'):
    """
    This function calls the .query() method on a DataFrame
    and additionally computes the ratio of resulting rows
    over the original number of rows.

    The result is a tuple with the filtered dataframe as first
    element and the filter ratio as second element.
    """
    df_out = df_in.query(query)
    ratio = df_out.shape[0] / df_in.shape[0]
    print('{} = {:.2f} %'.format(ratio_name, 100 * ratio))
    return df_out, ratio


def remove_outliers(df_in, column, sigma=3):
    """
    Very simple filter that removes outlier rows
    from a DataFrame based on the distance from the
    mean value measured in standard deviations.
    """
    return df_in[np.abs(df_in[column] - df_in[column].mean()) <= (sigma * df_in[column].std())]


def only_outliers(df_in, column, sigma=3):
    """
    Very simple filter that only keeps outlier rows
    from a DataFrame based on the distance from the
    mean value measured in standard deviations.
    """
    return df_in[np.abs(df_in[column] - df_in[column].mean()) > (sigma * df_in[column].std())]


def move_col_to_beginning_of_df(df_in, colname):
    """
    Small helper to move a column to the beginning of the DataFrame
    """
    cols = df_in.columns.tolist()
    cols.insert(0, cols.pop(cols.index(colname)))
    return df_in.reindex(columns=cols)


def compare_df_cols(df_list, col_list, mode=1):
    """
    Helper to compare the values of common columns between different dataframes

    Mode 1: iterate over columns as top level and DataFrames as second level
    Mode 2: iterate over DataFrames as top level and columns as second level
    """
    if mode == 1:
        colstoconcat = [df.loc[:, col].rename(df.loc[:, col].name + '_' + str(i + 1))
                        for col in col_list
                        for i, df in enumerate(df_list)]
    elif mode == 2:
        colstoconcat = [df.loc[:, col].rename(df.loc[:, col].name + '_' + str(i + 1))
                        for i, df in enumerate(df_list)
                        for col in col_list]
    else:
        return None

    return pd.concat(colstoconcat, axis=1)


def mad(c=None, name='mad'):
    try:
        import statsmodels.robust as smrb
    except ImportError:  # pragma: no cover
        raise ImportError(
            'The statsmodels package is required in order to use this function '
            'but it will not installed automatically by the zkpytb package. '
            'Please install it yourself.'
        )

    if c is not None:
        def _mad(x):
            return smrb.mad(x, c=c)
    else:
        def _mad(x):
            return smrb.mad(x)
    _mad.__name__ = name
    return _mad


def percentile(n):

    def _percentile(x):
        return np.percentile(x, n)
    _percentile.__name__ = 'percentile_%02d' % n

    return _percentile


def describe_numeric_1d(series):
    """
    Patched version of pandas' .describe() function for Series
    which includes the calculation of the median absolute deviation and interquartile range
    """
    stat_index = (['count', 'mean', 'std', 'mad', 'mad_c1', 'iqr', 'min']
                  + pd.io.formats.format.format_percentiles(extended_percentiles) + ['max'])
    d = (
        [
            series.count(),
            series.mean(),
            series.std(),
            mad()(series.dropna()),
            mad(1, name='mad_c1')(series.dropna()),
            scipy.stats.iqr(series, nan_policy='omit'),
            series.min()
        ]
        + series.quantile(extended_percentiles).tolist()
        + [series.max()]
    )
    return pd.Series(d, index=stat_index, name=series.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for the `pandas` module of `zkpytb` package."""

import pytest
from pytest import approx, raises

np = pytest.importorskip("numpy")  # noqa
pd = pytest.importorskip("pandas")  # noqa

from numpy.testing import assert_array_equal

from zkpytb.pandas import (
    extended_percentiles,
    compare_df_cols,
    df_query_with_ratio,
    move_col_to_beginning_of_df,
    only_outliers,
    remove_outliers,
    tdescr,
    mad,
    percentile,
    describe_numeric_1d,
)


describe_numeric_1d_expected_col_list = [
    'count', 'mean', 'std', 'mad', 'mad_c1', 'iqr', 'min', '1%', '5%', '25%', '50%', '75%', '95%', '99%', 'max'
]


def dummyfunc():
    return


@pytest.fixture(scope='module',
                params=[None, 0, 1, 3.14, set(), 'string', True, dict(), list(), dummyfunc],
                ids=['none', '0', '1', 'float', 'set', 'string', 'bool', 'dict', 'list', 'func'])
def not_a_df(request):
    return request.param


def generate_test_df(df_len, rand_seed):
    return pd.DataFrame({
        'a': np.zeros(df_len),
        'b': np.ones(df_len),
        'c': np.arange(df_len) + 1,
        'd': np.arange(df_len, dtype=float),
        'e': np.random.RandomState(rand_seed).rand(df_len),
        'f': np.random.RandomState(rand_seed).randn(df_len),
        'g': np.random.RandomState(rand_seed).random_integers(0, 100, df_len),
        'n': np.zeros(df_len) + np.nan,
    }, index=range(df_len))


@pytest.fixture(scope='module')
def df1():
    return generate_test_df(df_len=10, rand_seed=123456)


@pytest.fixture(scope='module')
def df2():
    return generate_test_df(df_len=10, rand_seed=998877)


def test_tdescr_not_a_df(not_a_df):
    with raises(AttributeError, match="has no attribute 'describe'"):
        tdescr(not_a_df)


def test_df_query_with_ratio_not_a_df(not_a_df):
    with raises(AttributeError, match="has no attribute 'query'"):
        df_query_with_ratio(not_a_df, '')


@pytest.fixture(params=[None, [.1, .25, .5, .75, .9]])
def percentiles(request):
    return request.param


@pytest.fixture(params=[True, False])
def disp(request):
    return request.param


def test_tdescr_df1(df1, percentiles, disp):
    res = tdescr(df1, percentiles=percentiles, disp=disp)
    assert isinstance(res, pd.DataFrame)
    res_rows, res_cols = res.shape
    epl = len(extended_percentiles) if percentiles is None else len(percentiles)
    assert res_cols == 5 + epl
    assert res_rows == len(df1.columns)


def test_df_query_with_ratio(df1):
    res, ratio = df_query_with_ratio(df1, 'c > 5')
    assert ratio == approx(0.5)
    assert_array_equal(res.index, np.arange(5, 10))


def test_remove_outliers(df1):
    res = remove_outliers(df1, 'f', 1)
    assert_array_equal(res.index, [1, 2, 3, 5, 6, 7, 8])


def test_only_outliers(df1):
    res = only_outliers(df1, 'f', 1)
    assert_array_equal(res.index, [0, 4, 9])


def test_move_col_to_beginning_of_df(df1):
    res = move_col_to_beginning_of_df(df1, 'e')
    assert_array_equal(res.columns, ['e', 'a', 'b', 'c', 'd', 'f', 'g', 'n'])


def test_compare_df_cols_mode1(df1, df2):
    res = compare_df_cols([df1, df2], ['e', 'f', 'g'], mode=1)
    assert list(res.columns) == ['e_1', 'e_2', 'f_1', 'f_2', 'g_1', 'g_2']


def test_compare_df_cols_mode2(df1, df2):
    res = compare_df_cols([df1, df2], ['e', 'f', 'g'], mode=2)
    assert list(res.columns) == ['e_1', 'f_1', 'g_1', 'e_2', 'f_2', 'g_2']


def test_compare_df_cols_mode3(df1, df2):
    res = compare_df_cols([df1, df2], ['e', 'f', 'g'], mode=3)
    assert res is None


def test_mad_df1(df1):
    f = mad()
    assert f.__name__ == 'mad'
    assert f(df1.a) == pytest.approx(0.0)
    assert f(df1.b) == pytest.approx(0.0)
    assert f(df1.c) == pytest.approx(3.7065055462640051)
    assert f(df1.d) == pytest.approx(3.7065055462640051)
    assert f(df1.e) == pytest.approx(0.32667683062127778)
    assert f(df1.f) == pytest.approx(0.93021508625544813)
    assert f(df1.g) == pytest.approx(17.791226622067224)


def test_mad_c1_df1(df1):
    f = mad(c=1, name='mad_c1')
    assert f.__name__ == 'mad_c1'
    assert f(df1.a) == pytest.approx(0.0)
    assert f(df1.b) == pytest.approx(0.0)
    assert f(df1.c) == pytest.approx(2.5)
    assert f(df1.d) == pytest.approx(2.5)
    assert f(df1.e) == pytest.approx(0.22034017388059335)
    assert f(df1.f) == pytest.approx(0.6274205411570638)
    assert f(df1.g) == pytest.approx(12.0)


def test_percentile_df1(df1):
    f = percentile(25)
    assert f.__name__ == 'percentile_25'
    assert f(df1.a) == pytest.approx(0.0)
    assert f(df1.b) == pytest.approx(1.0)
    assert f(df1.c) == pytest.approx(3.25)
    assert f(df1.d) == pytest.approx(2.25)
    assert f(df1.e) == pytest.approx(0.27941244048295077)
    assert f(df1.f) == pytest.approx(-1.1127832698328843)
    assert f(df1.g) == pytest.approx(37.75)


def test_describe_numeric_1d_df1(df1):
    res = {c: describe_numeric_1d(df1[c]) for c in df1.columns if c != 'n'}

    # column a: all zeros
    assert list(res['a'].index) == describe_numeric_1d_expected_col_list
    assert(res['a']['count']) == 10
    for c in ['1%', '5%', '25%', '50%', '75%', '95%', '99%', 'iqr', 'mad', 'mad_c1', 'max', 'mean', 'min', 'std']:
        assert res['a'][c] == pytest.approx(0.0)

    # column b: all ones
    assert list(res['b'].index) == describe_numeric_1d_expected_col_list
    assert(res['b']['count']) == 10
    for c in ['1%', '5%', '25%', '50%', '75%', '95%', '99%', 'max', 'mean', 'min']:
        assert res['b'][c] == pytest.approx(1.0)
    for c in ['iqr', 'mad', 'mad_c1', 'std']:
        assert res['b'][c] == pytest.approx(0.0)

    # column c: 1 to 10
    assert list(res['c'].index) == describe_numeric_1d_expected_col_list
    assert(res['c']['count']) == 10
    assert res['c']['mean'] == pytest.approx(5.5)
    assert res['c']['std'] == pytest.approx(3.027650)
    assert res['c']['mad'] == pytest.approx(3.706506)
    assert res['c']['mad_c1'] == pytest.approx(2.5)
    assert res['c']['iqr'] == pytest.approx(4.5)
    assert res['c']['min'] == pytest.approx(1.0)
    assert res['c']['1%'] == pytest.approx(1.09)
    assert res['c']['5%'] == pytest.approx(1.45)
    assert res['c']['25%'] == pytest.approx(3.25)
    assert res['c']['50%'] == pytest.approx(5.5)
    assert res['c']['75%'] == pytest.approx(7.75)
    assert res['c']['95%'] == pytest.approx(9.55)
    assert res['c']['99%'] == pytest.approx(9.91)
    assert res['c']['max'] == pytest.approx(10.0)



"""
Helper functions related to json

Author: Marc Gallet
"""


import datetime
import decimal
import json
import uuid
import pathlib


class JSONEncoder(json.JSONEncoder):
    """
    A custom JSONEncoder that can handle a bit more data types than the one from stdlib.
    """
    def default(self, o):
        # early passthrough if it works by default
        try:
            return json.JSONEncoder.default(self, o)
        except Exception:
            pass
        # handle Path objects
        if isinstance(o, pathlib.Path):
            return str(o).replace('\\', '/')
        # handle UUID objects
        if isinstance(o, uuid.UUID):
            return str(o)
        if isinstance(o, (datetime.datetime, datetime.time, datetime.date)):
            return o.isoformat()
        if isinstance(o, datetime.timedelta):
            return o.total_seconds()
        if isinstance(o, (complex, decimal.Decimal)):
            return str(o)
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = parse_requirements('requirements.txt')
test_requirements = parse_requirements('requirements_dev.txt')

setup(
    name='zkpytb',
    version='0.1.1',
    description="A collection of useful functions by Zertrin",
    long_description=readme + '\n\n' + history,
    author="Marc Gallet",
    author_email='zertrin@gmail.com',
    url='https://github.com/zertrin/zkpytb',
    project_urls={
        'Documentation': 'https://zkpytb.readthedocs.io/',
        'Say Thanks!': 'https://saythanks.io/to/zertrin',
        'Source': 'https://github.com/zertrin/zkpytb',
        'Tracker': 'https://github.com/zertrin/zkpytb/issues',
    },
    packages=find_packages(include=['zkpytb']),
    include_package_data=True,
    install_requires=requirements,
    python_requires='>=3.4, <4',
    license="MIT license",
    zip_safe=False,
    keywords='zkpytb',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    test_suite='tests',
    tests_require=test_requirements,
)



"""
Helper functions related to dictionaries

Author: Marc Gallet
"""


import hashlib
import json

from collections import OrderedDict

from zkpytb.json import JSONEncoder


class AutoDict(dict):
    """
    Default dict of dicts with infinite nesting (a.k.a. autovivification).
    """
    _base_class = dict

    def __getitem__(self, key):
        try:
            return self._base_class.__getitem__(self, key)
        except KeyError:
            value = self[key] = type(self)()
            return value

    def to_dict(self):
        """
        Convert this AutoDict to a normal dict (without autovivification).
        """
        return self._base_class(
            (key, (val.to_dict() if isinstance(val, AutoDict) else val))
            for key, val in self.items()
        )


class AutoOrderedDict(OrderedDict, AutoDict):
    """
    Default dict of OrderedDicts with infinite nesting (a.k.a. autovivification).
    """
    _base_class = OrderedDict


def filter_dict_callfunc(dict_in, func):
    assert(isinstance(dict_in, dict))
    assert(callable(func))
    return {k: v for (k, v) in dict_in.items() if func(k, v)}


def filter_dict_only_scalar_values(dict_in):
    assert(isinstance(dict_in, dict))
    return {k: v for (k, v) in dict_in.items() if not hasattr(v, '__iter__')}


def filter_dict_with_keylist(dict_in, keylist, blacklistmode=False):
    assert(isinstance(dict_in, dict))
    assert(isinstance(keylist, list))
    if blacklistmode:
        return {k: v for (k, v) in dict_in.items() if k not in keylist}
    else:
        return {k: v for (k, v) in dict_in.items() if k in keylist}


def mergedicts(dict1, dict2):
    assert(isinstance(dict1, dict))
    assert(isinstance(dict2, dict))
    for k in set(dict1.keys()).union(dict2.keys()):
        if k in dict1 and k in dict2:
            if isinstance(dict1[k], dict) and isinstance(dict2[k], dict):
                yield (k, dict(mergedicts(dict1[k], dict2[k])))
            else:
                # If one of the values is not a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                yield (k, dict2[k])
                # Alternatively, replace this with exception raiser to alert you of value conflicts
        elif k in dict1:
            yield (k, dict1[k])
        else:
            yield (k, dict2[k])


def dict_stable_json_repr(dict_in):
    return json.dumps(dict_in, sort_keys=True, cls=JSONEncoder)


def hashdict(dict_in, method='sha1'):
    assert(isinstance(dict_in, dict))
    h = hashlib.new(method)
    dict_repr = dict_stable_json_repr(dict_in)
    h.update(dict_repr.encode('utf-8'))
    return h.hexdigest()


def dict_values_map(f, d):
    """
    Simple helper to apply a function to the values of a dictionary.

    This is similar to built-in map() but only accept one iterable dict.
    """
    return {k: f(v) for k, v in d.items()}
